use std::collections::hash_map::Entry;
use std::collections::{HashMap, HashSet};
use std::path::Path;

use color_eyre::eyre::{eyre, ContextCompat, WrapErr};
use color_eyre::Result;
use image::RgbaImage;
use resvg::usvg::{TextRendering, TreeParsing, TreeTextToPath};
use tiny_skia::{Pixmap, PixmapMut, Transform};

use crate::model::config::{IconFormat, IconPack};
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::model::image_filter::ImageFilter;

pub fn load_icons(
    icon_packs_by_id: HashMap<String, IconPack>,
    descriptors: HashSet<IconDescriptor>,
    key_size: (u16, u16),
    dpi: f32,
) -> Result<HashMap<IconDescriptor, Pixmap>> {
    let mut original_images_by_source: HashMap<IconDescriptorSource, RgbaImage> = HashMap::new();
    let mut icons_by_descriptor: HashMap<IconDescriptor, Pixmap> = HashMap::new();
    let mut fonts_db = resvg::usvg::fontdb::Database::new();
    fonts_db.load_system_fonts();

    for descriptor in descriptors {
        if descriptor.source == IconDescriptorSource::None {
            continue;
        }

        let image = match original_images_by_source.entry(descriptor.source.clone()) {
            Entry::Occupied(o) => o.into_mut(),
            Entry::Vacant(v) => v.insert(read_image(&icon_packs_by_id, dpi, &fonts_db, descriptor.source)?),
        };

        dbg!(image.height());
    }

    Ok(icons_by_descriptor)
}

fn read_image(
    icon_packs_by_id: &HashMap<String, IconPack>,
    dpi: f32,
    fonts_db: &resvg::usvg::fontdb::Database,
    source: IconDescriptorSource,
) -> Result<RgbaImage> {
    let path = match source {
        IconDescriptorSource::None => unreachable!(),
        IconDescriptorSource::Path(path) => path,
        IconDescriptorSource::IconPack { pack_id, icon_id } => {
            let pack = icon_packs_by_id.get(&pack_id).wrap_err_with(|| format!("Unknown icon pack: @{}", pack_id))?;

            let extension = match pack.format {
                IconFormat::PNG => "png",
                IconFormat::SVG => "svg",
            };

            pack.path.join(icon_id + "." + extension)
        }
    };

    Ok(match path.extension() {
        None => return Err(eyre!("Invalid icon path: {:?}", path)),
        Some(extension) => match extension.to_string_lossy().as_ref() {
            "png" => image::open(path.clone())
                .wrap_err_with(|| format!("Failed to open or decode the PNG file at {}", path.to_string_lossy()))?
                .into_rgba8(),
            "svg" => {
                read_image_from_svg(&path, dpi, fonts_db).wrap_err_with(|| format!("Failed to open or decode the SVG file at {}", path.to_string_lossy()))?
            }
            extension => return Err(eyre!("Invalid file extension, only *.png and *.svg are allowed: {}", extension)),
        },
    })
}

fn read_image_from_svg(path: &Path, dpi: f32, font_db: &resvg::usvg::fontdb::Database) -> Result<RgbaImage> {
    let data = std::fs::read(path)?;
    let mut tree = resvg::usvg::Tree::from_data(
        &data,
        &resvg::usvg::Options {
            dpi,
            font_family: "Inter".to_owned(),
            font_size: 11.0,
            text_rendering: TextRendering::OptimizeLegibility,
            ..resvg::usvg::Options::default()
        },
    )?;

    tree.convert_text(font_db);

    let render_tree = resvg::Tree::from_usvg(&tree);
    let width = render_tree.size.width().ceil() as u32;
    let height = render_tree.size.height().ceil() as u32;
    let mut pixmap = Pixmap::new(width, height).unwrap();
    let mut pixmap_mut: PixmapMut = pixmap.as_mut();

    render_tree.render(Transform::identity(), &mut pixmap_mut);

    Ok(RgbaImage::from_vec(
        width,
        height,
        pixmap
            .pixels_mut()
            .iter()
            .flat_map(|p| {
                let p = p.demultiply();
                [p.red(), p.green(), p.blue(), p.alpha()]
            })
            .collect(),
    )
    .unwrap())
}

fn apply_filter(original: &RgbaImage, filter: &ImageFilter) -> RgbaImage {
    let mut result = original.clone();

    if filter.alpha != 1.0 {
        for pixel in result.pixels_mut() {
            pixel.0[3] = (pixel.0[3] as f32 * filter.alpha).floor() as u8;
        }
    }

    result
}
