use std::collections::HashMap;
use std::sync::Arc;
use std::thread;

use color_eyre::eyre::{ContextCompat, WrapErr};
use color_eyre::Result;
use enum_map::EnumMap;
use enum_ordinalize::Ordinalize;
use flume::{Receiver, Sender};
use log::{debug, trace};
use rgb::RGB8;

use loupedeck_serial::characteristics::{LoupedeckButton, LoupedeckDeviceKeyGridCharacteristics};
use loupedeck_serial::commands::VibrationPattern;
use loupedeck_serial::device::LoupedeckDevice;
use loupedeck_serial::events::LoupedeckEvent;

use crate::model;
use crate::model::icon_descriptor::IconDescriptor;
use crate::model::{ButtonPosition, KeyPath, KeyPosition, KnobPath};
use crate::runner::graphics::render_key;
use crate::runner::state::{Key, State, StateChangeCommand};

mod graphics;
mod state;

pub async fn start(config: model::config::Config) -> Result<()> {
    let config = Arc::new(config);
    let device = LoupedeckDevice::discover()?
        .first()
        .wrap_err("No device connected.")?
        .connect()
        .wrap_err("Connecting to the device failed.")?;

    device.set_brightness(0.0);
    device.vibrate(VibrationPattern::RiseFall);

    let events_receiver = device.events();
    let (commands_sender, commands_receiver) = flume::bounded::<StateChangeCommand>(20);

    commands_sender
        .send(StateChangeCommand::SetActivePages {
            knob_page_id: config.initial.knob_page.clone(),
            key_page_id: config.initial.key_page.clone(),
        })
        .unwrap();

    let cloned_config = Arc::clone(&config);
    let cloned_commands_sender = commands_sender.clone();
    let io_worker_thread = thread::Builder::new()
        .name("deckster IO worker".to_owned())
        .spawn(move || do_io_work(cloned_config, device, events_receiver, cloned_commands_sender, commands_receiver))
        .wrap_err("Could not spawn the worker thread")?;

    io_worker_thread.join().unwrap();

    Ok(())
}

fn create_state(config: &model::config::Config) -> State {
    let key_pages_by_id: HashMap<_, _> = config
        .key_pages_by_id
        .iter()
        .map(|(id, p)| state::KeyPage {
            id: id.clone(),
            keys_by_position: p
                .keys
                .iter()
                .map(|(position, k)| Key {
                    path: KeyPath {
                        page_id: p.id.clone(),
                        position: *position,
                    },
                    label: k.label.clone().unwrap_or_default(),
                    icon: IconDescriptor::default(),
                })
                .map(|k| (k.path.position, k))
                .collect(),
        })
        .map(|p| (p.id.clone(), p))
        .collect();

    let knob_pages_by_id: HashMap<_, _> = config
        .knob_pages_by_id
        .iter()
        .map(|(id, p)| state::KnobPage {
            id: id.clone(),
            knobs_by_position: EnumMap::from_fn(|position| {
                let knob_config = &p.knobs[position];

                state::Knob {
                    path: KnobPath {
                        page_id: p.id.clone(),
                        position,
                    },
                    label: knob_config.label.clone(),
                    icon: knob_config.icon.0.clone(),
                    value: 0.0,
                }
            }),
        })
        .map(|p| (p.id.clone(), p))
        .collect();

    State {
        active_key_page_id: config.initial.key_page.clone(),
        active_knob_page_id: config.initial.knob_page.clone(),
        key_pages_by_id,
        knob_pages_by_id,
    }
}

enum IoWork {
    Event(LoupedeckEvent),
    Command(StateChangeCommand),
}

fn do_io_work(
    config: Arc<model::config::Config>,
    device: LoupedeckDevice,
    events_receiver: Receiver<LoupedeckEvent>,
    commands_sender: Sender<StateChangeCommand>,
    commands_receiver: Receiver<StateChangeCommand>,
) {
    let mut state = create_state(&config);

    loop {
        let a = flume::Selector::new()
            .recv(&events_receiver, |e| IoWork::Event(e.unwrap()))
            .recv(&commands_receiver, |c| IoWork::Command(c.unwrap()))
            .wait();

        match a {
            IoWork::Event(event) => {
                if !handle_event(&config, &state, &device, &commands_sender, event) {
                    break;
                }
            }
            IoWork::Command(command) => handle_command(&config, &mut state, &device, command),
        }
    }
}

fn handle_event(
    config: &model::config::Config,
    state: &State,
    device: &LoupedeckDevice,
    commands_sender: &Sender<StateChangeCommand>,
    event: LoupedeckEvent,
) -> bool {
    trace!("Handling event: {:?}", &event);

    match event {
        LoupedeckEvent::Disconnected => return false,
        LoupedeckEvent::ButtonDown { button } => {
            let position = ButtonPosition::of(&button);
            let button_config = &config.buttons[position];

            commands_sender
                .send(StateChangeCommand::SetActivePages {
                    key_page_id: button_config.key_page.as_ref().unwrap_or(&state.active_key_page_id).clone(),
                    knob_page_id: button_config.knob_page.as_ref().unwrap_or(&state.active_knob_page_id).clone(),
                })
                .unwrap()
        }
        LoupedeckEvent::Touch { x, y, is_end, .. } => {
            if is_end {
                let characteristics = device.characteristics();
                let display = characteristics.get_display_at_coordinates(x, y);

                if let Some(display) = display {
                    if display.name == characteristics.key_grid.display.name {
                        let key_index = characteristics.key_grid.get_key_at_global_coordinates(x, y);
                        if let Some(key_index) = key_index {
                            let position = KeyPosition {
                                x: (key_index % characteristics.key_grid.columns) as u16 + 1,
                                y: (key_index / characteristics.key_grid.columns) as u16 + 1,
                            };

                            let path = KeyPath {
                                page_id: state.active_key_page_id.clone(),
                                position,
                            };

                            let value = if state.get_key(&path).label.is_empty() {
                                "a".to_owned()
                            } else {
                                String::new()
                            };

                            commands_sender.send(StateChangeCommand::SetKeyLabel { path, value }).unwrap();
                        }
                    }
                }
            }
        }
        _ => {}
    }

    true
}

fn handle_command(config: &model::config::Config, state: &mut State, device: &LoupedeckDevice, command: StateChangeCommand) {
    debug!("Handling command: {:?}", &command);

    match command {
        StateChangeCommand::SetActivePages { key_page_id, knob_page_id } => {
            state.active_key_page_id = key_page_id;
            state.active_knob_page_id = knob_page_id;

            for button in LoupedeckButton::VARIANTS {
                let position = ButtonPosition::of(button);

                device.set_button_color(*button, get_correct_button_color(config, state, position)).unwrap();
            }

            let key_grid = &device.characteristics().key_grid;
            for index in 0..(key_grid.rows * key_grid.columns) {
                draw_key_at_index(device, state, index);
            }

            device.refresh_display(&key_grid.display).unwrap();
        }
        StateChangeCommand::SetKeyLabel { path, value } => {
            state.mutate_key_for_command(
                "SetKeyLabel",
                &path,
                Box::new(|k| {
                    k.label = value;
                }),
            );

            draw_key_at_path_if_visible(device, state, path);
            device.refresh_display(&device.characteristics().key_grid.display).unwrap();
        }
        StateChangeCommand::SetKeyIcon { path, value } => {
            state.mutate_key_for_command(
                "SetKeyIcon",
                &path,
                Box::new(|k| {
                    k.icon = value;
                }),
            );

            draw_key_at_path_if_visible(device, state, path);
            device.refresh_display(&device.characteristics().key_grid.display).unwrap();
        }
    }
}

// active -> config.active_button_color
// no actions defined -> #000000
// inactive -> config.inactive_button_color
fn get_correct_button_color(config: &model::config::Config, state: &mut State, button_position: ButtonPosition) -> RGB8 {
    let button_config = &config.buttons[button_position];

    if let Some(key_page) = &button_config.key_page {
        if key_page == &state.active_key_page_id {
            if let Some(knob_page) = &button_config.knob_page {
                if knob_page == &state.active_knob_page_id {
                    return config.active_button_color.0;
                }
            }
        }
    } else if button_config.knob_page.is_none() {
        return RGB8::new(0, 0, 0);
    }

    config.inactive_button_color.0
}

fn get_key_index_for_position(key_grid: &LoupedeckDeviceKeyGridCharacteristics, position: KeyPosition) -> Option<u8> {
    if (position.x - 1) >= key_grid.columns as u16 || (position.y - 1) >= key_grid.rows as u16 {
        None
    } else {
        let x = (position.x - 1) as u8;
        let y = (position.y - 1) as u8;
        Some(y * key_grid.columns + x)
    }
}

fn get_key_position_for_index(key_grid: &LoupedeckDeviceKeyGridCharacteristics, index: u8) -> KeyPosition {
    let x = index % key_grid.columns;
    let y = index / key_grid.columns;

    KeyPosition {
        x: (x + 1) as u16,
        y: (y + 1) as u16,
    }
}

fn draw_key(device: &LoupedeckDevice, key_grid: &LoupedeckDeviceKeyGridCharacteristics, index: u8, key: Option<&Key>) {
    let (x, y, w, h) = key_grid.get_local_key_rect_xywh(index).unwrap();

    device
        .replace_framebuffer_area_raw(&key_grid.display, x, y, w, h, render_key(w, h, key_grid.display.endianness, key))
        .unwrap();
}

fn draw_key_at_index(device: &LoupedeckDevice, state: &State, index: u8) {
    let key_grid = &device.characteristics().key_grid;
    let position = get_key_position_for_index(key_grid, index);

    draw_key(device, key_grid, index, state.active_key_page().keys_by_position.get(&position));
}

fn draw_key_at_position_if_visible(device: &LoupedeckDevice, state: &State, position: KeyPosition) {
    let key_grid = &device.characteristics().key_grid;
    let index = get_key_index_for_position(key_grid, position);

    if let Some(index) = index {
        draw_key(device, key_grid, index, state.active_key_page().keys_by_position.get(&position));
    }
}

fn draw_key_at_path_if_visible(device: &LoupedeckDevice, state: &State, path: KeyPath) {
    if state.active_key_page_id == path.page_id {
        draw_key_at_position_if_visible(device, state, path.position);
    }
}
